# -*- coding: utf-8 -*-
import abc
from typing import List

from bkflow_dmn.exception import ValidationError, HitPolicyMatchError


class BaseHitPolicy(metaclass=abc.ABCMeta):
    """
    Abstract base class for all policies.
    """

    MAYBE_MULTIPLE_OUTPUT = False
    EMPTY_OUTPUT = []

    def __init__(self, strict_mode=True, *args, **kwargs):
        self.strict_mode = strict_mode

    def __call__(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        if not isinstance(rule_results, list) or not isinstance(outputs, list) or len(rule_results) != len(outputs):
            raise ValidationError("rule_results and outputs must have the same length")
        return self.handle(rule_results, outputs, *args, **kwargs)

    @abc.abstractmethod
    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        pass

    def multiple_output(self):
        return self.MAYBE_MULTIPLE_OUTPUT


class UniqueHitPolicy(BaseHitPolicy):
    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        if self.strict_mode:
            if rule_results.count(True) != 1:
                raise HitPolicyMatchError("Unique Hit Policy requires exactly one True result")
            return outputs[rule_results.index(True)]
        try:
            return outputs[rule_results.index(True)]
        except ValueError:
            return self.EMPTY_OUTPUT


class AnyHitPolicy(BaseHitPolicy):
    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        if self.strict_mode:
            if not any(rule_results):
                raise HitPolicyMatchError("Any Hit Policy requires at least one True result")
            all_match_idxes = [i for i, r in enumerate(rule_results) if r is True]
            match_outputs = [outputs[i] for i in all_match_idxes]
            # 检查所有的 match_outputs 是否都相等
            if not all(output == match_outputs[0] for output in match_outputs):
                raise HitPolicyMatchError("All match outputs must be equal")
            return outputs[rule_results.index(True)]
        try:
            return outputs[rule_results.index(True)]
        except ValueError:
            return self.EMPTY_OUTPUT


class FirstHitPolicy(BaseHitPolicy):
    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        if self.strict_mode:
            if not any(rule_results):
                raise HitPolicyMatchError("First Hit Policy requires at least one True result")
            return outputs[rule_results.index(True)]
        try:
            return outputs[rule_results.index(True)]
        except ValueError:
            return self.EMPTY_OUTPUT


class PriorityHitPolicy(BaseHitPolicy):
    """
    如果不是 strict 模式，则第一个命中的最高优先级作为输出
    """

    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        if not any(rule_results):
            if self.strict_mode:
                raise HitPolicyMatchError("Priority Hit Policy requires at least one True result")
            return self.EMPTY_OUTPUT
        match_outputs = [outputs[i] for i in range(len(rule_results)) if rule_results[i]]
        match_outputs = sorted(match_outputs, key=lambda x: list(x[i] for i in range(len(x))), reverse=True)
        if self.strict_mode and len(match_outputs) > 1 and match_outputs[0] == match_outputs[1]:
            raise HitPolicyMatchError("Priority Hit Policy requires exactly one True result")
        return match_outputs[0]


class RuleOrderHitPolicy(BaseHitPolicy):
    MAYBE_MULTIPLE_OUTPUT = True

    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        return [outputs[i] for i in range(len(rule_results)) if rule_results[i]]


class OutputOrderHitPolicy(BaseHitPolicy):
    MAYBE_MULTIPLE_OUTPUT = True

    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        match_outputs = [outputs[i] for i in range(len(rule_results)) if rule_results[i]]
        return sorted(match_outputs, key=lambda x: list(x[i] for i in range(len(x))), reverse=True)


class CollectHitPolicy(BaseHitPolicy):
    MAYBE_MULTIPLE_OUTPUT = True
    COLLECT_METHODS = {
        "sum": sum,
        "max": max,
        "min": min,
        "count": len,
    }

    def __init__(self, strict_mode=True, *args, **kwargs):
        super().__init__(strict_mode=strict_mode, *args, **kwargs)
        self.collect_policy = kwargs.get("collect_policy", "").lower()

    def handle(self, rule_results: List[bool], outputs: List, *args, **kwargs):
        match_outputs = [outputs[i] for i in range(len(rule_results)) if rule_results[i]]
        if self.collect_policy in self.COLLECT_METHODS:
            method = self.COLLECT_METHODS[self.collect_policy]
            collect_result = [method(col) for col in zip(*match_outputs)]
            return collect_result
        return match_outputs

    def multiple_output(self):
        return self.collect_policy not in self.COLLECT_METHODS


def get_hit_policy(hit_policy_name: str, strict_mode=True, *args, **kwargs):
    """
    根据 hit_policy_name 获取对应的 hit_policy
    """
    mappings = {
        "Unique": UniqueHitPolicy,
        "Any": AnyHitPolicy,
        "First": FirstHitPolicy,
        "Priority": PriorityHitPolicy,
        "RuleOrder": RuleOrderHitPolicy,
        "OutputOrder": OutputOrderHitPolicy,
        "Collect": CollectHitPolicy,
    }
    if hit_policy_name != "Collect" and hit_policy_name.startswith("Collect"):
        collect_policy = hit_policy_name[len("Collect") + 1 : -1]
        return CollectHitPolicy(strict_mode=strict_mode, collect_policy=collect_policy, *args, **kwargs)
    else:
        return mappings[hit_policy_name](strict_mode, *args, **kwargs)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['bkflow_dmn']

package_data = \
{'': ['*']}

install_requires = \
['bkflow-feel >= 1.0.0,<2', 'pydantic <3']

extras_require = \
{'test': ['pytest >=7.0.1,<8', 'pytest-benchmark[histogram] >=3.4.1,<4']}

setup(name='bkflow-dmn',
      version='0.1.1',
      description='A DMN (Decision Model and Notation) Implementation for Python',
      author=None,
      author_email='normal-wls <weishi.swee@qq.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



# -*- coding: utf-8 -*-
from bkflow_dmn.data_model import SingleDecisionTable


def decide_single_table(decision_table: dict, facts: dict, strict_mode: bool = True):
    dt = SingleDecisionTable(**decision_table)
    result = dt.decide(facts, strict_mode)
    return result



# -*- coding: utf-8 -*-
from enum import Enum
from typing import List, Union

from pydantic import BaseModel, root_validator

from bkflow_feel.api import parse_expression

from bkflow_dmn.hit_policy import get_hit_policy


class HitPolicyEnum(str, Enum):
    Unique = "Unique"
    Any = "Any"
    First = "First"
    Priority = "Priority"
    RuleOrder = "RuleOrder"
    OutputOrder = "OutputOrder"
    Collect = "Collect"
    Collect_SUM = "Collect(Sum)"
    Collect_Max = "Collect(Max)"
    Collect_Min = "Collect(Min)"
    Collect_Count = "Collect(Count)"


class DataTableField(BaseModel):
    id: str
    name: str = None
    desc: str = None


class DataTable(BaseModel):
    cols: List[DataTableField]
    rows: List[Union[List[str], str]]

    @root_validator(skip_on_failure=True)
    def validate_length(cls, values: dict):
        for row in values["rows"]:
            if isinstance(row, list) and len(row) != len(values["cols"]):
                raise ValueError("the length of row should be the same as the length of cols")
        return values

    @property
    def col_ids(self):
        return [col.id for col in self.cols]


class SingleDecisionTable(BaseModel):
    title: str
    hit_policy: HitPolicyEnum = HitPolicyEnum.Unique
    inputs: DataTable
    outputs: DataTable

    def decide(self, facts: dict, strict_mode=True):
        """
        单表单次决策函数
        :param facts: 单次决策对应的事实，对应于决策上下文
        :param strict_mode: 是否采用严格模式，当为 True 时，决策逻辑会进行一些校验并在失败时抛出异常
        :return: 决策结果
        """
        feel_inputs: List[Union[List[str], str]] = self.feel_exp_of_inputs
        feel_outputs: List[str] = self.outputs.rows
        parsed_inputs = [
            all([parse_expression(feel_input, facts) for feel_input in feel_input_row])
            for feel_input_row in feel_inputs
        ]
        parsed_outputs = []
        for feel_output_row in feel_outputs:
            parsed_outputs.append([parse_expression(feel_output, facts) for feel_output in feel_output_row])
        hit_policy = get_hit_policy(self.hit_policy_value, strict_mode=strict_mode)
        outputs_result = hit_policy(parsed_inputs, parsed_outputs)
        final_result = []
        if hit_policy.multiple_output():
            final_result.extend(
                [
                    {key: value for key, value in zip(self.outputs.col_ids, outputs_result_row)}
                    for outputs_result_row in outputs_result
                ]
            )
        elif outputs_result:
            final_result.append({key: value for key, value in zip(self.outputs.col_ids, outputs_result)})
        return final_result

    @root_validator(skip_on_failure=True)
    def validate_length(cls, values: dict):
        if len(values["inputs"].rows) != len(values["outputs"].rows):
            raise ValueError("the length of inputs should be the same as the length of outputs")
        return values

    @property
    def hit_policy_value(self):
        return self.hit_policy.value

    @property
    def feel_exp_of_inputs(self):
        col_ids = [col.id for col in self.inputs.cols]
        operator_prefix = {
            ">": "",
            "<": "",
            "[": " in ",
            "(": " in ",
        }
        or_operator = ","

        result = []
        for row in self.inputs.rows:
            if isinstance(row, str):
                result.append([row])
                continue
            row_exps = []
            for idx, unit in enumerate(row):
                pured_unit = unit.strip()
                if pured_unit == "":
                    continue
                unit_exps = []
                vals = [val.strip() for val in pured_unit.split(or_operator) if val.strip()]
                for val in vals:
                    if val[0] in operator_prefix:
                        unit_exps.append(f"{col_ids[idx]}{operator_prefix[val[0]]}{val}")
                    else:
                        unit_exps.append(f"{col_ids[idx]}={val}")
                row_exps.append(f'({" or ".join(unit_exps)})' if len(unit_exps) > 1 else unit_exps[0])
            result.append(row_exps)
        return result
